specialCams = {}	-- Special camera panning
eventSounds = {}	-- Sounds made by the player, etc that can get an enemy's attention

local zUnit = 16 	-- Number of y pixels that correspond to each one Z block

lightsOn = true		-- Whether a room is lit
controlsDisabled = false	-- Controls may be disabled in some situations

-- Image that contains an image along with its file name
NamedImg = {}
function NamedImg:new(str)
	local o = {name = str, img = loadImg(str), hasName = true}
	setmetatable(o, self)
	self.__index = self

	o.w = o.img:getWidth()
	o.h = o.img:getHeight()

	return o
end

-- Image that has a shader associated with it
EffectImg = {}
function EffectImg:new(str, effect)
	local o = NamedImg:new(str)
	setmetatable(o, self)
	self.__index = self
	
	o.effect = effect
	o.hasEffect = true
	o.opacity = 1
	o.dOpacity = .98

	return o
end


-- Array of sound effects
soundArray = {}
-- Plays a single instance of a sound
function playSound(name)
	if soundArray[name] then
		love.audio.play(soundArray[name])
		return true
	end
	local s = love.audio.newSource(name..".mp3", "static")
	love.audio.play(s)
end

-- Draws sprite without regard to the camera
function drawNoCam(obj)
	love.graphics.draw(obj.img, obj.x, obj.y)
end

function degToRad(d)
	return d * (math.pi/180)
end

-- Takes an object name and returns its proper image
function loadImg(str)
	return love.graphics.newImage("Images/"..str..".png")
end

function distToObj(from, to)
	return math.sqrt(math.pow(from.x - to.x, 2) + math.pow(from.y - to.y, 2))
end

function angleToObj(from, to)
	return math.atan2(to.y-from.y, to.x-from.x)
end

function onScreen(obj)
	if obj.x + obj.w >= camRect.x and obj.x <= camRect:right() and obj.y + obj.h >= camRect.y and
	 obj.y <= camRect:bottom() then
		return true
	end
	return false
end

function nearScreen(obj, d)
	if obj.x + obj.w >= camRect.x - d and obj.x <= camRect:right() + d and obj.y + obj.h >= camRect.y - d and
	 obj.y <= camRect:bottom() + d then
		return true
	end
	return false
end

-- Checks if a number is within a certain range (n) of another
function approx(x, y, n)
	if math.abs(x-y) <= n then
		return true
	end
	return false
end


-- For shapes
Vertex = {}
function Vertex:new(x, y, z)
	local o = {x = x, y = y, z = z}
	setmetatable(o, self)
	self.__index = self
	return o
end

-- Line
Line = {}
function Line:new(x1, y1, x2, y2)
	local o = {x1 = x1, y1 = y1, x2 = x2, y2 = y2}
	setmetatable(o, self)
	self.__index = self
	return o
end
function Line:intersectRect(line, rect)
	if rect.w < 1 then
		return false
	end
	return self:intersectLine(line, Line:new(rect.x, rect.y, rect:right(), rect.y)) or 
		self:intersectLine(line, Line:new(rect:right(), rect.y, rect:right(), rect:bottom())) or
		self:intersectLine(line, Line:new(rect:right(), rect:bottom(), rect.x, rect:bottom())) or
		self:intersectLine(line, Line:new(rect.x, rect:bottom(), rect.x, rect.y))
end
function Line:intersectLine(l1, l2)
	local q = (l1.y1 - l2.y1) * (l2.x2 - l2.x1) - (l1.x1 - l2.x1) * (l2.y2 - l2.y1)
	local d = (l1.x2 - l1.x1) * (l2.y2 - l2.y1) - (l1.y2 - l1.y1) * (l2.x2 - l2.x1)
	
	if d == 0 then
		return false
	end
	
	local r = q/d

	q = (l1.y1 - l2.y1) * (l1.x2 - l1.x1) - (l1.x1 - l2.x1) * (l1.y2 - l1.y1)

	local s = q/d

	if (r < 0 or r > 1 or s < 0 or s > 1) then
		return false
	end

	return true
end

--***************--
-- Simple rect class
Rect = {}
function Rect:new(x, y, w, h)
	local o = {x = x, y = y, z = 0, w = w, h = h, dx = 0, dy = 0, dz = 0, color = {255, 255, 255, 255}, isSlope = false,
	drawOffsetX = 0, drawOffsetY = 0, life = 100000, died = false, takesDamage = true, delete = false, depth = 1,
	gunDamageVal = 0, animTime = 0, currFrame = 1, maxAngle = math.pi/3, alerted = false, tookDamage = false, pushDx = 0,
	pushDy = 0}
	setmetatable(o, self)
	self.__index = self
	return o
end
function Rect:interact()
end
function Rect:action()
end
function Rect:move()
	self.x = self.x + self.dx
	self.y = self.y + self.dy
end
function Rect:moveAnim()
end
function Rect:enableAnimation()
	self.animated = true
	-- Determine whether it's a special animation and get the image
	if self.currAnimation.timer ~= nil then
		self.img = self.currAnimation[1]
	else
		self.img = self.currAnimation[1][1]
	end
end
function Rect:draw()
	local tr = trans(self)
	if self.img ~= nil then
		-- For named images
		if self.img.hasName then
			if self.img.hasEffect then
				self.img.effect(self)
				love.graphics.draw(self.img.img, math.floor(tr.x + self.drawOffsetX), math.floor(tr.y + self.drawOffsetY))
				previousEffect()
			else
				love.graphics.draw(self.img.img, math.floor(tr.x + self.drawOffsetX), math.floor(tr.y + self.drawOffsetY))
			end			
		-- For unnamed images
		else
			love.graphics.draw(self.img, math.floor(tr.x + self.drawOffsetX), math.floor(tr.y + self.drawOffsetY))
		end
	else
		if self.color[4] < 5 then
			return false
		end
		love.graphics.setColor(self.color)
		love.graphics.rectangle("fill", math.floor(tr.x ), math.floor(tr.y ), self.w, self.h)
		love.graphics.setColor({255, 255, 255, 255})
	end
end
function Rect:right()
	return self.x + self.w
end
function Rect:bottom()
	return self.y + self.h
end
function Rect:centerX()
	return self.x + self.w/2
end
-- Checks if something is within a certain field of view
function Rect:withinMaxAngle(maxAngle)
	local angle = math.atan2(player.y - self.y, player.x - self.x)
	if self.facing == 1 and (angle > maxAngle or angle < -maxAngle) then
		return false
	elseif self.facing == -1 and ((angle > 0 and angle < math.pi - maxAngle) or (angle < 0 and angle > -math.pi + maxAngle)) then
		return false
	end
	return true
end
-- Changes an object's animation and starts it from the beginning
function Rect:setAnim(anim)
	self.currAnimation = anim
	self.currFrame = 1
	self.animTime = 0
	-- Check whether it's a special animation and get the first frame
	if self.currAnimation.timer ~= nil then
		self.img = self.currAnimation[self.currFrame]
	else
		self.img = self.currAnimation[self.currFrame][1]
	end
	-- Since we can sometimes change an animation's timing, reset it here
	if self.oldTimer then
		self.currAnimation.timer = self.oldTimer
	end
end
-- Death action for an object
function Rect:death()
	self.delete = true
end
-- Animates a sprite at a set framerate
function Rect:animate()
	-- Increments animation timer and advances frame if >= current animation time
	self.animTime = self.animTime + 1
	if self.animTime >= self.currAnimation.timer then
		self.currFrame = self.currFrame + 1
		self.animTime = 0
		if self.currFrame > #self.currAnimation then
			if self.currAnimation.noLoop then
				self.currFrame = self.currFrame - 1
				self.animTime = 100
			else
				self.currFrame = 1
			end
		end
		self.img = self.currAnimation[self.currFrame]
	end
end
-- Animates with a variable frame rate
-- {{frame1, time}, {frame2, time}, ...}
-- To use this function, override an object's animate function to call this
-- function self:animate() self:specialAnimate() end
function Rect:specialAnimate()
	self.animTime = self.animTime + 1
	if self.animTime >= self.currAnimation[self.currFrame][2] then
		self.currFrame = self.currFrame + 1
		self.animTime = 0
		if self.currFrame > #self.currAnimation then
			if self.currAnimation.noLoop then
				self.currFrame = self.currFrame - 1
				self.animTime = 100
			else
				self.currFrame = 1
			end
		end
		self.img = self.currAnimation[self.currFrame][1]
	end
end

-- Special draw function for 3D areas. 
function draw3D(self)
	local drawOffsetY = self.drawOffsetY + self.z * zUnit
	local tr = trans(self)
	if self.img ~= nil then
		-- For named images
		if self.img.hasName then
			if self.img.hasEffect then
				self.img.effect(self)
				love.graphics.draw(self.img.img, math.floor(tr.x + self.drawOffsetX), math.floor(tr.y + drawOffsetY))
				previousEffect()
			else
				love.graphics.draw(self.img.img, math.floor(tr.x + self.drawOffsetX), math.floor(tr.y + drawOffsetY))
			end			
		-- For unnamed images
		else
			love.graphics.draw(self.img, math.floor(tr.x + self.drawOffsetX), math.floor(tr.y + drawOffsetY))
		end
	else
		if self.color[4] < 5 then
			return false
		end
		love.graphics.setColor(self.color)
		love.graphics.rectangle("fill", math.floor(tr.x ), math.floor(tr.y ), self.w, self.h)
		love.graphics.setColor({255, 255, 255, 255})
	end
end


-- Block that allows you to grab on or jump up through
OneWay = Rect:new(0, 0, 0, 0)
function OneWay:new(x, y, w, h)
	local o = Rect:new(x, y, w, h)
	o.oneWay = true
	setmetatable(o, self)
	self.__index = self

	return o
end

--*****************--
-- A rect with an angle/slope
-- Note: slopes side lengths needs to perfectly fit the angle
-- e.g.: 45 degree slope needs a perfect square rect
Slope = Rect:new()
function Slope:new(x, y, w, h, angle)
	local o = Rect:new(x, y, w, h)
	setmetatable(o, self)
	self.__index = self

	o.angle = angle
	o.isSlope = true
	o.drawOffsetX = 0
	return o
end
-- Returns the y position an object should be at
function Slope:getY(obj)
	-- We'll check for overlap with either the left corner of the foot
	-- or the right corner of the foot, based on slope angle

	-- Angles going up
	if self.angle > 0 then
		local x = obj.footRect.x - self.x
		x = x + obj.footRect.w
		return -x * math.tan(self.angle) + self:bottom() + 2
	end
	-- Angles going down
	local x = math.abs(obj.footRect.x - self:right())
	local y = - x * math.tan(-self.angle) + self:bottom() + 2
	return y
end
-- Check to see whether an object is below the slope of the object
function Slope:colSlope(obj)
	local y = self:getY(obj)
	if obj:bottom() >= y then
		return true
	end
	return false
end

-- Portal to another area
Door = Rect:new()
function Door:new(x, y, w, h, dest, name)
	local o = {life = 1000, x = x, y = y, w = w, h = h, isSlope = false, isDoor = true, destination = dest, interactive = true}
	setmetatable(o, self)
	self.__index = self
	if name ~= nil then
		o.img = loadImg(name)
	end
	return o
end
function Door:interact()
	loadMap(self.destination)
end

-- Door that's removed upon spawn
OneWayDoor = Door:new()
function OneWayDoor:new(x, y, dest)
	local o = Door:new(x, y, 0, 0, dest)
	setmetatable(o, self)
	self.__index = self
	return o
end
function OneWayDoor:action()
	self.delete = true
end


entities = {}
--****************--
-- Entity capable of grabbing onto ledges; includes the player
Human = Rect:new()
function Human:new(x, y, w, h, xOff)
	local xOff = xOff or 0 -- The amount of overhang for grabbing onto ledges
	local o = {x = x, y = y, w = w, h = h, rect = Rect:new(x, y, w, h), grabRect = Rect:new(x, y, w, 6), dx = 0, dy = 0,
	 footRect = Rect:new(x, y+h, w, 3), xOffset = xOff, onGround = true, onLedge = false, facing = -1, onSlope = false,
	 crouchVal = 11, interactive = false, maxLife = 10, meleeDamageVal = 2, fallVel = 0, gravity = .15,
	 animated = false, animTime = 0, currFrame = 1, takesDamage = true, friendlyFireOn = false, crouchV = 8,
	 hasTwoFeet = false, lookTimer = 0, maxLookTimer = 5, exDx = 0, exDy = 0, maxLookDist = screenWidth / 2,
	 soundTimer = 0, maxSoundTimer = 5, soundThreshold = 60, maxInvestigateTimer = 100, investigateTimer = 0}
	o.life = o.maxLife
	o.spawn = o

	
	o.totalMoveTimer = 0
	o.meleeRect = Rect:new(x, y, 10, 10)
	o.meleeRect.xOffset = 0
	o.meleeRect.yOffset = 10

	-- States relating to melee attacks
	o.meleeTriggered = false			-- Button pressed to cause a melee attack
	o.meleeDx = 7 						-- Melee attack speed
	o.startedMelee = false				-- Melee is in motion
	o.totalMelee = 0					-- Distance covered in attack
	o.maxMelee = 35						-- Max range
	o.reachedMaxMelee = false			-- Attack returning to player to end
	o.meleeRecovery = 0					-- Countdown to when the player can next use a melee attack

	-- Keeps track of enemies hit so they don't take continuous damage
	o.enemiesHit = {}

	o.grabRect.color = {255, 0, 0, 255}
	o.footRect.color = {0, 255, 0, 255}
	setmetatable(o, self)
	self.__index = self
	return o
end
-- Where the feet should be located
function Human:getFootY()
	return self.y + self.h
end
-- Spawns feet
function Human:spawnFeet()
	self.hasTwoFeet = true
	self.leftFoot = Rect:new(self.x, self:bottom(), 3, 4)
	self.rightFoot = Rect:new(self:right() - 3, self:bottom(), 3, 4)
end
-- Updates foot position
function Human:updateFeet()
	self.leftFoot.x = self.x
	self.leftFoot.y = self:bottom()
	self.rightFoot.x = self:right() - 3
	self.rightFoot.y = self:bottom()
end
-- move() moves the main body and all related subunits of the entity
function Human:move()
	-- Move main body, then have subrects move in relation
	moveHuman(self, walls)
	-- grabRect is for hanging onto ledges
	self.grabRect.x = self.x - self.facing*self.xOffset
	self.grabRect.y = self.y
	-- footRect determines whether the player is on a platform
	self.footRect.x = self.x
	self.footRect.y = self:getFootY()
	-- Fall due to gravity
	if not self.onGround and self.dy < 4 then
		self.dy = self.dy + self.gravity
	end
	-- meleeRect handles melee attack collision
	self.meleeRect.x = self.x + (self.w / 2) - (self.meleeRect.w / 2) + self.meleeRect.xOffset
	self.meleeRect.y = self.y + self.meleeRect.yOffset

	self.meleeRecovery = self.meleeRecovery - 1
end
-- Moves with momentum and takes time to turn around
function Human:moveMomentum()
	-- Maintain a little momentum when letting go unless crouched
	if self.direction == 0 and not self.crouched then
		-- If the player's been walking for a long period, maintain some momentum
		if self.totalMoveTimer > 50 then
			self.dx = self.dx * .8
			if math.abs(self.dx) < .2 then
				self.dx = 0
				self.totalMoveTimer = 0
			end
		-- Otherwise, let them stop immediately
		else
			self.dx = 0
		end
	end

	-- If not crouched, have the player slide a bit when turning
	if not self.crouched and self.onGround then
		if self.direction < 0 and self.dx > 0 then
			self.dx = self.dx - self.s / self.crouchV
		elseif self.direction < 0 and self.dx <= 0 then
			self.dx = -self.s
		elseif self.direction > 0 and self.dx < 0 then
			self.dx = self.dx + self.s / self.crouchV
		elseif self.direction > 0 and self.dx >= 0 then
			self.dx = self.s
		end
	else
		self.dx = self.s * self.direction
	end
	-- Move main body, then have subrects move in relation
	moveHuman(self, walls)
	-- grabRect is for hanging onto ledges
	self.grabRect.x = self.x - self.facing*self.xOffset
	self.grabRect.y = self.y
	-- footRect determines whether the player is on a platform
	self.footRect.x = self.x
	self.footRect.y = self.y + self.h
	if not self.onGround and self.dy < 4 then
		self.dy = self.dy + self.gravity
	end
	-- meleeRect handles melee attack collision
	self.meleeRect.x = self.x + (self.w / 2) - (self.meleeRect.w / 2) + self.meleeRect.xOffset
	self.meleeRect.y = self.y + self.meleeRect.yOffset

	self.meleeRecovery = self.meleeRecovery - 1
end
-- Listens for any nearby sounds
function Human:hear()
	self.soundTimer = self.soundTimer + 1
	if self.soundTimer >= self.maxSoundTimer then
		self.soundTimer = 0
		local center = {x = self.x + self.w/2, y = self.y + 4}
		-- Check for nearby sounds and keep a note of where they are to investigate
		for i = 1, #eventSounds do
			local dist = distToObj({x = eventSounds[i].x, y = eventSounds[i].y * 2}, center)
			if dist < 380 and eventSounds[i].vol / dist >= self.soundThreshold then
				-- If not currently investigating, start doing so
				if self.heardSound == false then
					self.heardSound = true
					self.investigateTimer = 0
					-- Show question mark icon above head
					self.showQuestion = true
					self.currQuestion = 1
				-- If already snooping, only increase time by a max of up to half the max time
				else
					if self.investigateTimer > self.maxInvestigateTimer / 2 then
						self.investigateTimer = self.maxInvestigateTimer / 2
					end
				end

				self.soundLocation = {x = eventSounds[i].x, y = eventSounds[i].y}
			end
		end
	end
end
function Human:updateQuestion()
	if self.showQuestion then
		if self.investigateTimer > 2 then
			self.currQuestion = 2
		end
		if self.investigateTimer > 4 then
			self.currQuestion = 3
		end
		if self.investigateTimer > 70 then
			self.showQuestion = false
		end
		self.investigateTimer = self.investigateTimer + 1
	end
end
-- Looks for a direct line to a specified object
function Human:look(obj)
	local pathBlocked = false
	local collided = false
	-- if you collided with the enemy, automatically  trigger a response
	if collideRect(self, obj) then
		collided = true
	else
		-- Only check during regular intervals; not every frame
		self.lookTimer = self.lookTimer + 1
		if self.lookTimer >= self.maxLookTimer then
			self.lookTimer = 0
		else
			return false
		end

		-- Only test if facing towards the targer
		if (obj.x > self.x and self.facing == -1) or (obj.x < self.x and self.facing == 1) then
			return false
		end

		-- If the distance from the target to the seer is too great, ignore
		if distToObj(self, obj) > self.maxLookDist then
			return false
		end

		-- Check the angle in relation to the target; looking straight up or down is weird
		if not self:withinMaxAngle(self.maxAngle) then
			return false
		end
		
		-- Determine whether to look for the target's left or right side
		local x = nil
		if obj:right() < self.x then
			x = obj:right()
		else
			x = obj.x
		end

		-- If hanging and above the seer, we want to check for feet
		local y = obj.y
		if obj.onLedge and obj.y < self.y then
			y = obj:bottom()
		end

		-- Get target's position and a list of all rects contained within the space
		-- between the guard's position and the player's
		local line = Line:new(self.x + self.w/2, self.y + 3, x, y)

		for i = 1, #walls do
			pathBlocked = Line:intersectRect(line, walls[i])
			if pathBlocked then
				break
			end
		end
	end
	-- If a clear line of sight is established, register as true
	if collided or not pathBlocked then
		self.heardSound = false

		return true
	end
	return false
end
-- Jumps up at a set velocity
function Human:jump()
	if self.onGround then
		self.inJump = true
		self.onGround = false
		self.dy = -4.25
		-- If crouched, then uncrouch
		if self.crawling then
			self:crawl()
		elseif self.crouched then
			self:crouch()
		end
	end
end
-- Crouches, reducing size
function Human:crouch()
	-- Get into crouch position
	if not self.crouched then
		self.y = self.y + self.crouchVal
		self.h = self.h - self.crouchVal
		self.drawOffsetY = self.drawOffsetY - self.crouchVal
		self.crouched = true
	-- Uncrouch, if possible
	else
		self.y = self.y - self.crouchVal
		self.h = self.h + self.crouchVal
		self.drawOffsetY = self.drawOffsetY + self.crouchVal
		self.crouched = false

		-- Check and see if the obj's head is now colliding with any blocks, and undo if so
		for i = 1, #walls do
			if collideRect(walls[i], self) and walls[i].y < self.footRect.y and walls[i].w > 0 then
				self.y = self.y + self.crouchVal
				self.h = self.h - self.crouchVal
				self.drawOffsetY = self.drawOffsetY - self.crouchVal
				self.crouched = true
				break
			end
		end
	end
end
-- Launches a projectile from the character at a given angle and velocity
function Human:shoot()
	-- Where the bullet comes from
	local origin_x = self.x + self.w/2
	local origin_y = self.y + self.h/2

	local facing = self.facing
	if self.onLedge then
		facing = facing * -1
	end

	-- Speed and angle
	local s = -10 * facing
	local ang = math.pi

	-- Fire the bullet
	fireProjectile(origin_x, origin_y, 6, 6, s, ang, 0, 0, self.gunDamageVal, self)
end
-- Melee attacks launch a small "boomerang" rect out of the player to attack
function Human:melee()
	-- When hanging on a wall, we want the obj to shoot away from it
	local facing = self.facing
	if self.onLedge == true then
		facing = facing * -1
	end
	-- Perform an attack that launches a rect 
	if self.meleeRecovery < 1 then
		-- Launch out
		if not self.reachedMaxMelee then
			self.meleeRect.xOffset = (math.abs(self.meleeRect.xOffset) + self.meleeDx) * facing
			self.totalMelee = self.totalMelee + self.meleeDx
			if self.totalMelee >= self.maxMelee then
				self.reachedMaxMelee = true
			end
		-- Pull inward
		else
			self.meleeRect.xOffset = (math.abs(self.meleeRect.xOffset) - self.meleeDx) * facing
			self.totalMelee = self.totalMelee - self.meleeDx
			if self.totalMelee <= 0 then
				self.reachedMaxMelee = false
				self.meleeTriggered = false
				self.meleeRecovery = 10
				-- Reset the damaged table
				self.enemiesHit = {}
			end
		end
	end
	-- Damage enemies
	if self.meleeTriggered then
		for i = 1, #entities do
			if entities[i] ~= self and collideRect(self.meleeRect, entities[i]) then
				-- When we hit an enemy, we add a reference to a table to ensure it's
				-- not hit again until the attack resets
				if self.enemiesHit[entities[i]] == nil then
					entities[i].life = entities[i].life - self.meleeDamageVal
					self.enemiesHit[entities[i]] = 1
				end
			end
		end
	end
end
-- Action of an object; includes animation, attack, etc
function Human:action()

end
-- Dying
function Human:death()
	self.delete = true
	shatter(self)
end

-- Fires a projectile in a direction
function fireProjectile(origin_x, origin_y, w, h, s, ang, dS, dAng, damageVal, spawn)
	local p = Rect:new(origin_x, origin_y, w, h)
	p.dx = math.cos(ang) * s
	p.dy = math.sin(ang) * s
	p.img = loadImg("bullet")
	p.damageVal = damageVal
	p.spawn = spawn

	table.insert(bullets, p)
end


-- I'm going to try a very rudimentary implementation of a rasterizer
local oldDrawPoly = [[
function drawPoly(poly, width, color)
	local width = width or 1
	local color = color or {255, 0, 0, 500}
	love.graphics.setColor(color)
	love.graphics.setLineWidth(width)
	for i = 1, table.getn(poly)-1, 1 do
		x1 = (poly[i+1].x/poly[i+1].z) * 200
		x2 = (poly[i].x/poly[i].z) * 200
		y1 = (poly[i+1].y/poly[i+1].z) * 200
		y2 = (poly[i].y/poly[i].z) * 200
		love.graphics.line(x1+400, y1+400, x2+400, y2+400)
	end

	-- Link beginning to end
	local len = table.getn(poly)
	local x1 = (poly[1].x/poly[1].z) * 200
	local x2 = (poly[len].x/poly[len].z) * 200
	local y1 = (poly[1].y/poly[1].z) * 200
	local y2 = (poly[len].y/poly[len].z) * 200
	love.graphics.line(x1+400, y1+400, x2+400, y2+400)
	
	love.graphics.setColor(255, 255, 255, 1000)
end
]]

-- I'm going to try a very rudimentary implementation of a rasterizer
function drawPoly(poly, width, color)
	local center = {x = camRect.x + screenWidth/2, y = camRect.y + screenWidth/2}

	local width = width or 1
	local color = color or {255, 0, 0, 500}
	love.graphics.setColor(color)
	love.graphics.setLineWidth(width)

	-- Note: we now need to center this on the screen

	for i = 1, table.getn(poly)-1, 1 do
		local x1 = (poly[i+1].x/poly[i+1].z)
		local x2 = (poly[i].x/poly[i].z)
		local y1 = (poly[i+1].y/poly[i+1].z)
		local y2 = (poly[i].y/poly[i].z)
		love.graphics.line(x1, y1, x2, y2)
	end

	-- Link beginning to end
	local len = table.getn(poly)
	local x1 = (poly[1].x/poly[1].z)
	local x2 = (poly[len].x/poly[len].z)
	local y1 = (poly[1].y/poly[1].z)
	local y2 = (poly[len].y/poly[len].z)
	love.graphics.line(x1, y1, x2, y2)
	
	love.graphics.setColor(255, 255, 255, 255)
end

function movePoly(poly, dx, dy, dz)
	for i = 1, table.getn(poly), 1 do
		poly[i].x = poly[i].x + dx
		poly[i].y = poly[i].y + dy
		poly[i].z = poly[i].z + dz
	end
end

-- Tests for collisions between any rect objects
function collideRect(rect1, rect2)
	if rect2.y < rect1.y+rect1.h and rect2.y+rect2.h > rect1.y and rect2.x < rect1.x+rect1.w and rect2.x+rect2.w > rect1.x then
		return true
	end
	return false
end

-- Tests for collisions between 3D blocks
function collideBlock(rect1, rect2)
	if rect2.y < rect1.y+rect1.h and rect2.y+rect2.h > rect1.y and rect2.x < rect1.x+rect1.w and rect2.x+rect2.w > rect1.x then
		if rect1.z < rect2.z + rect2.depth and rect1.z + rect1.depth > rect2.z then
			return true
		end
	end
	return false
end

-- Movement of objects and basic wall collision
walls = {}
bgWalls = {}
fgWalls = {}
bullets = {}
function moveObj(obj, walls)
	local colWith = nil

	if obj.footRect ~= nil then
		return moveHuman(obj, walls)
	end

	obj.x = obj.x + obj.dx
	for i = 1, table.getn(walls), 1 do
		if collideRect(obj, walls[i]) then
			local colWith = walls[i]
			-- Place the object at the proper place in relation to the wall
			if obj.dx > 0 then
				local overlap = (obj.x + obj.w) - walls[i].x
				obj.x = obj.x - overlap
			else
				local overlap = (walls[i].x + walls[i].w) - obj.x
				obj.x = obj.x + overlap
			end
		end
	end
	obj.y = obj.y + obj.dy
	for i = 1, table.getn(walls), 1 do
		if collideRect(obj, walls[i]) then
			local colWith = walls[i]
			-- Place the object at the proper place in relation to the wall
			if obj.dy > 0 then
				local overlap = (obj.y + obj.h) - walls[i].y
				obj.y = obj.y - overlap
				obj.onGround = true
			else
				local overlap = (walls[i].y + walls[i].h) - obj.y
				obj.y = obj.y + overlap
			end
		end
	end
end

-- Movement of objects with hand/foot rects
function moveHuman(obj, walls)
	-- If the obj isn't on a platform or hanging, fall
	obj.footRect.collided = false
	if obj.hasTwoFeet then
		obj.leftFoot.collided = false
		obj.rightFoot.collided = false
	end
	local touchedSlope = false

	-- If the hands contact a platform above, we know there's no ledge
	local touchedFromAbove = false
	local touchedLedge = false
	local ledge = nil

	obj.landed = false

	obj.oldDx = obj.dx
	obj.dx = obj.dx + obj.exDx
	obj.exDx = 0
	
	local oldX = obj.x
	local xCol = false
	obj.xCol = false
	-- Crouched units move slower
	if obj.crouched and not obj.crawling then
		obj.x = obj.x + obj.dx/obj.crouchV
	elseif obj.crawling then
		obj.x = obj.x + obj.dx / obj.crawlV
	else
		obj.x = obj.x + obj.dx
	end


	if obj.w == 0 then
		return false
	end

	-- X movement
	for i = 1, table.getn(walls), 1 do
		if collideRect(obj, walls[i]) and walls[i].w > 0 and not walls[i].oneWay then
			obj.colWith = walls[i]
			-- When moving horizontally w/ slopes, we need to properly slide down them
			if walls[i].isSlope	then
				if (obj.dx < 0 and walls[i].angle > 0) or (obj.dx > 0 and walls[i].angle < 0) then
					local y = walls[i]:getY(obj)
					if y < walls[i].y then
						y = walls[i].y
					end
					obj.y = y - obj.h
					obj.onGround = true
					touchedSlope = true
				end

			-- Handle non-slopes normally
			else
				local colWith = walls[i]
				-- Place the object at the proper place in relation to the wall
				if obj.dx > 0 then
					local overlap = (obj.x + obj.w) - walls[i].x
					obj.x = obj.x - overlap
				else
					local overlap = (walls[i].x + walls[i].w) - obj.x
					obj.x = obj.x + overlap
				end
				obj.xCol = true
				walls[i].colWith = obj
			end
		end
		-- Check whether feet collide with any tiles
		if obj.hasTwoFeet == false then
			if collideRect(obj.footRect, walls[i]) and not walls[i].isSlope and walls[i].w > 0 then
				obj.footRect.collided = true
			end
		-- If we have two feet, we test differently
		else
			if collideRect(obj.leftFoot, walls[i]) and not walls[i].isSlope and walls[i].w > 0 then
				obj.leftFoot.collided = true
				obj.footRect.collided = true
			end
			if collideRect(obj.rightFoot, walls[i]) and not walls[i].isSlope and walls[i].w > 0 then
				obj.rightFoot.collided = true
				obj.footRect.collided = true
			end
		end
	end
	-- Set the DX to the former value
	obj.dx = obj.oldDx

	-- If there's nothing below the human, fall
	if obj.footRect.collided == false then
		obj.onGround = false
	end
	
	obj.dy = obj.dy + obj.exDy 		-- Special effects acting upon objects movement (wind, etc)
	obj.exDy = 0

	-- Y movement
	local oldY = obj.y
	obj.y = obj.y + obj.dy
	local yCol = false
	obj.yCol = false

	obj.oldDy = obj.dy
	obj.dy = obj.dy + obj.exDy
	obj.exDy = 0
	
	-- Y movement
	for i = 1, table.getn(walls), 1 do
		if collideRect(obj, walls[i]) and walls[i].w > 0 then
			local colWith = walls[i]
			obj.colWith = walls[i]
			-- Place the object at the proper place in relation to the wall
			-- Moving downwards
			if obj.dy >= 0 and (not walls[i].oneWay or (walls[i].oneWay and obj:bottom() < walls[i]:bottom())) then
				-- Collision involving slopes is a bit different; handle appropriately
				if walls[i].isSlope then
					if walls[i]:colSlope(obj) then
						-- Sets position that obj should be at on slope depending on x pos
						local y = walls[i]:getY(obj)
						if y < walls[i].y then
							y = walls[i].y
						end
						obj.y = y - obj.h
						obj.onGround = true
						touchedSlope = true
						obj.inJump = false
					end
				-- If it's one-way, only go on top if the foot is aligned with the top
				elseif walls[i].oneWay then
					-- Some objects should disregard OneWay blocks
					if not obj.noOneWay then
						if obj:bottom() <= walls[i].y + 4 then
							walls[i].steppedOn = true
							local overlap = (obj.y + obj.h) - walls[i].y
							obj.y = obj.y - overlap
							-- Make note of having landed after a fall
							if obj.onGround == false then
								obj.landed = true
								obj.fallVel = obj.dy
							end
							obj.onGround = true
							obj.dy = 0
							obj.yCol = true
							obj.inJump = false

							walls[i].colWith = obj
						end
					end
				-- Else, treat it normally
				else
					walls[i].steppedOn = true
					local overlap = (obj.y + obj.h) - walls[i].y
					obj.y = obj.y - overlap
					-- Make note of having landed after a fall
					if obj.onGround == false then
						obj.landed = true
						obj.fallVel = obj.dy
					end
					obj.onGround = true
					obj.dy = 0
					obj.yCol = true
					obj.inJump = false

					walls[i].colWith = obj
				end
			-- Moving upwards (usually jumping)
			elseif obj.dy < 0 and not walls[i].oneWay then
				-- Moving upwards and colliding with a slope:
				-- if hitting from the bottom, act as a normal wall
				-- if resting on top, ignore upward collision
				if walls[i].isSlope then
					if obj:bottom() > walls[i]:bottom() then
						local overlap = walls[i]:bottom() - obj.y
						obj.y = obj.y + overlap
					end
				else
					local overlap = (walls[i].y + walls[i].h) - obj.y
					obj.y = obj.y + overlap
					obj.yCol = true
				end
			end
		end
		-- Checks for collision with hands to grab onto ledges
		-- Only grabs onto corners of objects
		-- If we only want this to work on corners without blocks above,
		-- set a bool to determine if the obj collides with a rect above it
		-- If so, then keep falling
		if collideRect(obj.grabRect, walls[i]) and obj.dy > 0 and walls[i].w > 0 and obj.grabRect.w > 0 then
			if obj.grabRect.y <= walls[i].y then
				ledge = walls[i]
				touchedLedge = true	
			else
				touchedFromAbove = true
			end
		end
	end
	--obj.dy = oldDy

	-- Grab onto ledge if nothing above
	if touchedLedge and not touchedFromAbove then
		local collidedBelow = false
		local wasCrouched = false
		local wasCrawling = false
		
		-- If the object is crawling or crouched, check whether it'd collide below
		if obj.crouched or obj.crawling then
			if obj.crawling then
				--obj.h = obj.h + obj.crawlVal
				obj:crawl()
				wasCrawling = true
			elseif obj.crouched then
				--obj.h = obj.h + obj.crouchVal
				obj:crouch()
				wasCrouched = true
			end

			local oldY = obj.y
			local oldH = obj.h
			local oldDrawY = obj.drawOffsetY
			
			-- check if colliding with a block below, and if so, don't grab on
			for n = 1, #walls do
				obj.y = walls[n].y
				if collideRect(walls[n], obj) then
					collidedBelow = true
					obj.y = oldY
					obj.h = oldH
					obj.drawOffsetY = oldDrawY
					-- Get back into crawl/crouch
					if wasCrouched then
						--obj.h = obj.h - obj.crouchVal
						--obj:crouch()
					elseif wasCrawling then
						--obj.h = obj.h - obj.crawlVal
						--obj:crawl()
					end
					break
				end
			end
		end
		
		if not collidedBelow then
			obj.dy = 0
			obj.y = ledge.y
			obj.onGround = true
			obj.onLedge = true
		end
	end

	if oldX ~= obj.x or oldY ~= obj.y then
		obj.onLedge = false
	end
	if touchedSlope then
		obj.onSlope = true
	else
		obj.onSlope = false
	end
end

-- Movement of objects with hand/foot rects with additional Z movement
function moveHuman3D(obj, walls)
	-- If the obj isn't on a platform or hanging, fall
	obj.footRect.collided = false
	if obj.hasTwoFeet then
		obj.leftFoot.collided = false
		obj.rightFoot.collided = false
	end
	local touchedSlope = false

	-- If the hands contact a platform above, we know there's no ledge
	local touchedFromAbove = false
	local touchedLedge = false
	local ledge = nil

	obj.landed = false
	
	local oldX = obj.x
	local xCol = false
	obj.xCol = false
	-- Crouched units move slower
	if obj.crouched and not obj.crawling then
		obj.x = obj.x + obj.dx/obj.crouchV
	elseif obj.crawling then
		obj.x = obj.x + obj.dx / obj.crawlV
	else
		obj.x = obj.x + obj.dx
	end

	if obj.w == 0 then
		return false
	end

	obj.oldDx = obj.dx
	obj.dx = obj.dx + obj.exDx
	obj.exDx = 0

	-- X movement
	for i = 1, table.getn(walls), 1 do
		if collideBlock(obj, walls[i]) and walls[i].w > 0 and not walls[i].oneWay then
			-- When moving horizontally w/ slopes, we need to properly slide down them
			if walls[i].isSlope	then
				if (obj.dx < 0 and walls[i].angle > 0) or (obj.dx > 0 and walls[i].angle < 0) then
					local y = walls[i]:getY(obj)
					if y < walls[i].y then
						y = walls[i].y
					end
					obj.y = y - obj.h
					obj.onGround = true
					touchedSlope = true
				end

			-- Handle non-slopes normally
			else
				local colWith = walls[i]
				-- Place the object at the proper place in relation to the wall
				if obj.dx > 0 then
					local overlap = (obj.x + obj.w) - walls[i].x
					obj.x = obj.x - overlap
				else
					local overlap = (walls[i].x + walls[i].w) - obj.x
					obj.x = obj.x + overlap
				end
				obj.xCol = true
				walls[i].colWith = obj
			end
		end
		-- Check whether feet collide with any tiles
		if obj.hasTwoFeet == false then
			if collideRect(obj.footRect, walls[i]) and not walls[i].isSlope and walls[i].w > 0 then
				obj.footRect.collided = true
			end
		-- If we have two feet, we test differently
		else
			if collideRect(obj.leftFoot, walls[i]) and not walls[i].isSlope and walls[i].w > 0 then
				obj.leftFoot.collided = true
				obj.footRect.collided = true
			end
			if collideRect(obj.rightFoot, walls[i]) and not walls[i].isSlope and walls[i].w > 0 then
				obj.rightFoot.collided = true
				obj.footRect.collided = true
			end
		end
	end

	obj.dx = oldDx

	-- If there's nothing below the human, fall
	if obj.footRect.collided == false then
		obj.onGround = false
	end

	obj.dy = obj.dy + obj.exDy 		-- Special effects acting upon objects movement (wind, etc)
	obj.exDy = 0
	
	-- Y movement
	local oldY = obj.y
	obj.y = obj.y + obj.dy
	local yCol = false
	obj.yCol = false

	obj.oldDy = obj.dy
	obj.dy = obj.dy + obj.exDy
	obj.exDx = 0

	for i = 1, table.getn(walls), 1 do
		if collideBlock(obj, walls[i]) and walls[i].w > 0 then
			local colWith = walls[i]
			-- Place the object at the proper place in relation to the wall
			-- Moving downwards
			if obj.dy >= 0 and (not walls[i].oneWay or (walls[i].oneWay and obj:bottom() < walls[i]:bottom())) then
				-- Collision involving slopes is a bit different; handle appropriately
				if walls[i].isSlope then
					if walls[i]:colSlope(obj) then
						-- Sets position that obj should be at on slope depending on x pos
						local y = walls[i]:getY(obj)
						if y < walls[i].y then
							y = walls[i].y
						end
						obj.y = y - obj.h
						obj.onGround = true
						touchedSlope = true
						obj.inJump = false
					end
				-- If it's one-way, only go on top if the foot is aligned with the top
				elseif walls[i].oneWay then
					if obj:bottom() <= walls[i].y + 4 then
						walls[i].steppedOn = true
						local overlap = (obj.y + obj.h) - walls[i].y
						obj.y = obj.y - overlap
						-- Make note of having landed after a fall
						if obj.onGround == false then
							obj.landed = true
							obj.fallVel = obj.dy
						end
						obj.onGround = true
						obj.dy = 0
						obj.yCol = true
						obj.inJump = false

						walls[i].colWith = obj
					end
				-- Else, treat it normally
				else
					walls[i].steppedOn = true
					local overlap = (obj.y + obj.h) - walls[i].y
					obj.y = obj.y - overlap
					-- Make note of having landed after a fall
					if obj.onGround == false then
						obj.landed = true
						obj.fallVel = obj.dy
					end
					obj.onGround = true
					obj.dy = 0
					obj.yCol = true
					obj.inJump = false

					walls[i].colWith = obj
				end
			-- Moving upwards (usually jumping)
			elseif obj.dy < 0 and not walls[i].oneWay then
				-- Moving upwards and colliding with a slope:
				-- if hitting from the bottom, act as a normal wall
				-- if resting on top, ignore upward collision
				if walls[i].isSlope then
					if obj:bottom() > walls[i]:bottom() then
						local overlap = walls[i]:bottom() - obj.y
						obj.y = obj.y + overlap
					end
				else
					local overlap = (walls[i].y + walls[i].h) - obj.y
					obj.y = obj.y + overlap
					obj.yCol = true
				end
			end
		end
		-- Checks for collision with hands to grab onto ledges
		-- Only grabs onto corners of objects
		-- If we only want this to work on corners without blocks above,
		-- set a bool to determine if the obj collides with a rect above it
		-- If so, then keep falling
		if collideRect(obj.grabRect, walls[i]) and obj.dy > 0 and walls[i].w > 0 then
			if obj.grabRect.y <= walls[i].y then
				ledge = walls[i]
				touchedLedge = true	
			else
				touchedFromAbove = true
			end
		end
	end
	-- Grab onto ledge if nothing above
	if touchedLedge and not touchedFromAbove then
		local collidedBelow = false
		local wasCrouched = false
		local wasCrawling = false
		
		-- If the object is crawling or crouched, check whether it'd collide below
		if obj.crouched or obj.crawling then
			if obj.crawling then
				--obj.h = obj.h + obj.crawlVal
				obj:crawl()
				wasCrawling = true
			elseif obj.crouched then
				--obj.h = obj.h + obj.crouchVal
				obj:crouch()
				wasCrouched = true
			end

			local oldY = obj.y
			local oldH = obj.h
			local oldDrawY = obj.drawOffsetY
			
			-- check if colliding with a block below, and if so, don't grab on
			for n = 1, #walls do
				obj.y = walls[n].y
				if collideRect(walls[n], obj) then
					collidedBelow = true
					obj.y = oldY
					obj.h = oldH
					obj.drawOffsetY = oldDrawY
					-- Get back into crawl/crouch
					if wasCrouched then
						--obj.h = obj.h - obj.crouchVal
						--obj:crouch()
					elseif wasCrawling then
						--obj.h = obj.h - obj.crawlVal
						--obj:crawl()
					end
					break
				end
			end
		end
		
		if not collidedBelow then
			obj.dy = 0
			obj.y = ledge.y
			obj.onGround = true
			obj.onLedge = true
		end
	end

	obj.dy = oldDy

	if obj.exDz ~= nil then
		obj.dz = obj.dz + obj.exDz 		-- Special effects acting upon objects movement (wind, etc)
		obj.exDz = 0
	end

	-- Move on the Z axis
	if obj.dz ~= 0 then
		obj.z = obj.z + obj.dz

		for i = 1, #walls do
			local col = collideBlock(obj, walls[i])
			if col then
				if obj.dz < 0 then
					obj.z = walls[i].z + walls[i].depth
				elseif obj.dz > 0 then
					obj.z = walls[i].z - obj.depth
				end
			end
		end
	end

	if oldX ~= obj.x or oldY ~= obj.y then
		obj.onLedge = false
	end
	if touchedSlope then
		obj.onSlope = true
	else
		obj.onSlope = false
	end
end

-- Checks for left & right foot collisions with walls
function checkFeetCollision(rFoot, lFoot)
	local rFootCol = false
	local lFootCol = false

	for i = 1, #walls do
		if collideRect(walls[i], rFoot) then
			rFootCol = true
		end
		if collideRect(walls[i], lFoot) then
			lFootCol = true
		end
	end

	return {r = rFootCol, l = lFootCol}
end

-- Adjust the screen's rect to follow an object (usually the player) around
function adjustsubCamRect(obj, subCamRect)
	-- If we're in a situation with a special camera, pan and lock to it
	local colSpecialCam = nil
	for i = 1, #specialCams do
		-- Check for the player to be within a special camera, and if so, use it
		if collideRect(obj, specialCams[i]) then
			colSpecialCam = specialCams[i]
		end
	end

	if colSpecialCam ~= nil then
		return adjustSpecialCam(obj, colSpecialCam, subCamRect)
	end

	-- Otherwise, normal camera operations
	-- Move left
	if obj.x <= subCamRect.x then
		if obj.dx < 0 then
			subCamRect.x = subCamRect.x + obj.dx
		end
		if (subCamRect.x - obj.x) > 10 then
			subCamRect.x = subCamRect.x - 2
		end
	end
	-- Move right
	if obj:right() >= subCamRect:right() then
		if obj.dx > 0 then
			subCamRect.x = subCamRect.x + obj.dx
		end
		if (subCamRect:right() - obj:right()) < -10 then
			subCamRect.x = subCamRect.x + 2
		end
	end
	-- Move up
	if obj.y <= subCamRect.y then
		if obj.dy < 0 then
			subCamRect.y = subCamRect.y + obj.dy
		end
		if (subCamRect.y - obj.y) > 10 then
			subCamRect.y = subCamRect.y - 2
		end
	end
	-- Move down
	if obj:bottom() >= subCamRect:bottom() then
		if obj.dy > 0 then
			subCamRect.y = subCamRect.y + obj.dy
		end
		if (subCamRect:bottom() - obj:bottom()) < -10 then
			subCamRect.y = subCamRect.y + 2
		end
	end
	camRect.x = subCamRect.x - camOffsetX
	camRect.y = subCamRect.y - camOffsetY
end


-- Adjusts a camera to lock onto a location of interest
function adjustSpecialCam(obj, specialCam, subCamRect)
	local spc = Rect:new(specialCam.x + camOffsetX, specialCam.y + camOffsetY,
	 specialCam.w, specialCam.h)
	local scr = subCamRect

	-- Max distance something can be from the special rect
	local r = 2.5
	-- Speed at which camera pans
	local s = 2

	-- Handle situations where obj/player goes offscreen
	if obj.x < camRect.x then
		subCamRect.x = subCamRect.x - s
	elseif obj:right() > camRect:right() then
		subCamRect.x = subCamRect.x + s
	end
	if obj.y < camRect.y then
		subCamRect.y = subCamRect.y - s
	elseif obj:bottom() > camRect:bottom() then
		subCamRect.y = subCamRect.y + s
	end

	-- Move right/left
	if scr.x < spc.x and not approx(spc.x, scr.x, r) then
		subCamRect.x = subCamRect.x + s
	elseif scr:right() > spc:right() and not approx(scr:right(), spc:right(), r) then
		subCamRect.x = subCamRect.x - s
	end
	-- Move down/up
	if scr.y < spc.y and not approx(spc.y, scr.y, r) then
		subCamRect.y = subCamRect.y + s
	elseif scr:bottom() > spc:bottom() and not approx(scr:bottom(), spc:bottom(), r) then
		subCamRect.y = subCamRect.y - s
	end
	camRect.x = subCamRect.x - camOffsetX
	camRect.y = subCamRect.y - camOffsetY
end



-- Maintains a stack for graphical effects
effectStack = {}

-- Warps screen to have a geometric, drug-like effect
lsdRipple = 0
lsdIncr = 1
function lsdEffect()
	lsdRipple = lsdRipple + lsdIncr
	if lsdRipple > 20 then
		lsdIncr = -1
	elseif lsdRipple < -20 then
		lsdIncr = 1
	end
	--lsdRipple = lsdRipple % 20
	lsd:send("time", lsdRipple / 5)

	love.graphics.setPixelEffect(lsd)
	-- Pushes this effect to the stack
	table.insert(effectStack, lsdEffect)
end

-- Looks like a glitchy TV
tearShaderRipple = 0
tearShaderIncr = 1
tearIntensity = .1
function tearEffect()
	tearShaderRipple = tearShaderRipple + 1
	if tearShaderRipple > 40 then
		tearShaderRipple = -0
	end
	--lsdRipple = lsdRipple % 20
	if tearShaderRipple % 2 == 0 then
		tearShader:send("intensity", tearIntensity)
		tearShader:send("time", tearShaderRipple)
	end
	love.graphics.setPixelEffect(tearShader)
	-- Pushes this effect to the stack
	table.insert(effectStack, tearEffect)
end

-- Fills non-transparent areas with black
function blackEffect()
	love.graphics.setPixelEffect(blackShader)
	table.insert(effectStack, blackEffect)
end

-- Fills non-transparent areas with white
function whiteEffect()
	love.graphics.setPixelEffect(whiteShader)
	table.insert(effectStack, whiteEffect)
end

-- Darkens sprites
function darkEffect(val)
	local v = val or .5
	darkShader:send("intensity", v)
	love.graphics.setPixelEffect(darkShader)
	table.insert(effectStack, darkShader)
end

-- Light around an object
function circleLightEffect()
	love.graphics.setPixelEffect(circleLight)
	-- Pushes this effect to the stack
	table.insert(effectStack, circleLightEffect)
end

-- Displays a flickering gradient on sprites
-- Mostly useful for things like 
gradientFlicker = 0
function gradientEffect()
	gradientFlicker = gradientFlicker + 1
	gradientFlicker = gradientFlicker % 5
	gradient:send("time", gradientFlicker)
	love.graphics.setPixelEffect(gradient)
	-- Pushes this effect to the stack
	table.insert(effectStack, gradientEffect)
end

-- Warps and stretches an object
warpTimer = 1
warpIncr = .5
function warpEffect()
	warpTimer = warpTimer + warpIncr
	warpIncr  = warpIncr * .98
	if warpTimer > 20 then
		warpIncr = -1
	elseif warpTimer < -20 then
		warpIncr = 1
	end
	warp:send("time", warpTimer)
	love.graphics.setPixelEffect(warp)
	-- Pushes this effect to the stack
	table.insert(effectStack, warpEffect)
end

-- Makes an image transparent
function transparentEffect(obj)
	obj.img.opacity = obj.img.opacity * obj.img.dOpacity
	transparentShader:send("intensity", obj.img.opacity)
	love.graphics.setPixelEffect(transparentShader)

	table.insert(effectStack, transparentEffect)
end

-- Empties effectStack and resets the pixel effect
function resetEffects()
	effectStack = {}
	love.graphics.setPixelEffect()
end

-- Pops the current effect and then runs the previous one
function previousEffect()
	table.remove(effectStack, table.getn(effectStack))
	love.graphics.setPixelEffect(effectStack[#effectStack])
end

-- Translates object to display on screen
function trans(obj)
	return {x = obj.x - camRect.x, y = obj.y - camRect.y}
end
-- Gives a rect for translations
function transRect(obj)
	return {x = obj.x - camRect.x, y = obj.y - camRect.y, w = obj.w, h = obj.h}
end

-- Splits string into array
function splitString(s, pat)
	local pattern = pat
	if pattern == nil then
		pattern = "%S+"
	end
	words = {}
	for word in s:gmatch(pattern) do
		table.insert(words, word)
	end
	return words
end


textArray = {time = 100, text = {"Hello!"}}
textStack = {}
charArray = {} 				-- All character icons
-- Prepares text for dialogue with a timer
function createTextArray(_text, len)
	local textTime = 100
	-- Length of substring and max length before split
	local totalLen = 0
	local maxLen = len or 38

	local container = {}
	local subArray = {}
	local j = 1
	local text = splitString(_text)
	for i = 1, table.getn(text), 1 do
		totalLen = totalLen + string.len(text[i])
		textTime = textTime + 10

		if totalLen >= maxLen then
			subArray = table.concat(subArray, " ")
			container[j] = subArray
			j = j + 1
			subArray = {}
			totalLen = #text[i]
		end
		
		table.insert(subArray, text[i])

		-- If there's leftover text, just add it
		if (totalLen < maxLen) and (i >= #text) then
			subArray = table.concat(subArray, " ")
			container[j] = subArray
		end

		if textTime > 200 then
			textTime = 200
		end
	end
	return {time = textTime, text = container}
end
advanceText = false
-- Creates a text array with no timer, requiring a button press to advance
function createUntimedText(_text, len)
	local t = createTextArray(_text, len)
	t.isUntimed = true

	return t
end


--textAdvanceButton = loadImg("textadvancebutton")
-- Displays a dialogue frame if any dialogue exists
function displayDialogue(textStack)
	local character = nil
	local textArray = nil

	if textStack[1] ~= nil then
		character = textStack[1][1]
		textArray = textStack[1][2]
	end
	
	-- Loads next chunk of dialogue, if any, when time expires or button is pressed to advance
	if #textStack > 0 then
		if textArray.time <= 1 and not textArray.isUntimed then
			table.remove(textStack, 1)
		elseif textArray.isUntimed and advanceText then
			table.remove(textStack, 1)
		end
	else
		return false
	end
	
	-- Displays current text
	if textArray.time > 0 then
		gradientEffect()

		if charArray[character] ~= nil then
			love.graphics.draw(charArray[character], 14, 188)
		end
		love.graphics.draw(dialogueFrame, 0, 20)
		
		previousEffect()

		for i = 1, #textArray.text, 1 do
			love.graphics.print(textArray.text[i], 65, 195 + 15 * (i-1))
		end

		if not textArray.isUntimed then
			textArray.time = textArray.time - 1
		else
			love.graphics.draw(textAdvanceButton, screenWidth - 34, screenHeight - 18)
		end
	end
end

-- Creates a dialogue with text and character icon
function createDialogue(character, text)
	table.insert(textStack, {character, createUntimedText(text)})
end
--createDialogue("sed", "Hello, world! This is a very long string that takes two lines!")
--createDialogue("sed", "Here's more!")
-- Clears out the current dialogue stack
function clearDialogue()
	textStack = {}
end


-- Adds a "sound" triggered by some sort of action that can get enemy attention
function addEventSound(x, y, volume)
	local sound = {timer = 5, x = x, y = y, vol = volume}
	table.insert(eventSounds, sound)
end

-- Spawns explosions all over an object for a brief period of time, and then marks it for deletion
function explosionPeriod(self)
	if self.explodeTimer < self.maxExplodeTimer and self.explodeTimer % 3 == 0 then
		-- add explosion at random point
	else
		-- mark for deletion
	end
end


-- Splits image into rectangles and has them float off
function shatter(self)
	if self.died then
		return false
	end
	local imgData = love.image.newImageData("Images/"..self.img.name..".png")
	local rectangleArray = {}
	-- Note: imageData starts its arrays at 0, unlike Lua normally does at 1
	-- We end at -2 because we check the pixel to the right and below to build a rectangle
	for x = 2, imgData:getWidth() - 2 do
		for y = 0, imgData:getHeight() - 2 do
			if x % 2 == 0 and y % 2 == 0 then
				-- Get a pixel and only use those that are mostly opaque
				local r, g, b, a = imgData:getPixel(x, y)
				if a > 127 then
					local w = math.random() * 3
					local block = Rect:new(self.x + x - 1 + self.drawOffsetX, self.y + y - 1 + self.drawOffsetY, 2 + w, 2 + w)
					block.color = {r, g, b, a}
					block.dy = 1.2
					block.ddy = math.random() * -.1 - .05
					block.timer = 100
					block.action = function()
						block.dy = block.dy + block.ddy
						block.color[4] = block.color[4] - 2
						
						block.timer = block.timer - 1
						-- remove when alpha is less than 0
						if block.color[4] < 10 then	
							block.delete = true
						end
					end
					
					table.insert(entities, block)
				end
			end
		end
	end
	self.died = true
	self.img = nil
	-- If multiple images, clear them all
	if self.imgs ~= nil then
		for item in pairs(self.imgs) do
			self.imgs[item] = nil
		end
	end
	-- If any animations exist, clear them out
	if self.animationTable ~= nil then
		for item in pairs(self.animationTable) do
			self.animationTable[item] = nil
		end
	end
	self.animated = false
	self.color = {255, 255, 255, 255}
end

-- Alerts all entities of a presence
function alertAll(entities)
	for i = 1, #entities do
		entities[i].alerted = true
	end
end



-- Creates a mini text display above an object
miniTexts = {}
function addMiniText(str, obj)
	local m = MiniText:new(str, obj.x, obj.y)
	table.insert(miniTexts, m)
end

-- Tiny bit of text to display. Could be used for small item pickups, passing NPC comments, etc
arcadePi = love.graphics.newFont("ARCADEPI.TTF", 7.1)
MiniText = {}
function MiniText:new(_str, _x, _y)
	local o = {str = _str, x = _x, y = _y, timer = 170, maxTimer = 80, delete = false}
	setmetatable(o, self)
	self.__index = self

	return o
end
function MiniText:draw()
	love.graphics.setFont(arcadePi)
	self.timer = self.timer - 1
	if self.timer < 1 then
		self.delete = true
	end
	local r, g, b, a = love.graphics.getColor()
	local trX = self.x - camRect.x - 8
	local trY = self.y - camRect.y
	-- Print pink text in the back, then white in front
	-- Also, we'll want a period of full opacity before fading
	if self.timer > self.maxTimer then
		love.graphics.setColor(8, 20, 52, 255)
		love.graphics.print(self.str, trX+1, trY-1)
		love.graphics.setColor(58, 103, 195, 255)
		love.graphics.print(self.str, trX-1, trY+1)
		love.graphics.setColor(255, 255, 255, 255)
		love.graphics.print(self.str, trX, trY)
	else
		love.graphics.setColor(58, 103, 195, 255 - (self.maxTimer - self.timer)*3)
		love.graphics.print(self.str, trX-1, trY+1)
		love.graphics.setColor(255, 255, 255, 255 - (self.maxTimer - self.timer)*3)
		love.graphics.print(self.str, trX, trY)
	end

	love.graphics.setColor(r, g, b, a)
end


-- Camera related data
camOffsetX = 140
camOffsetY = 70
subCamRect = Rect:new(camOffsetX, camOffsetY, screenWidth - camOffsetX*2, screenHeight - camOffsetY*2)
subCamRect.color = {255, 255, 255, 255}
camRect = Rect:new(0, 0, screenWidth, screenHeight)

